/*
 * 3-D Helixpaint. Inspired by Electropaint.
 */

#include <stdlib.h>
#include <math.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include <gtk/gtkgl.h>
#include <GL/gl.h>

#include "helix.h"
#include "midi-source.h"
#include "midi-adjustment.h"

static HelixAttribs  attr_zero = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, {0.0, 0.0, 0.0}, 0 };

/* minimum / maximum / value */
static gdouble helix_defaults[HELIX_SLOT_LAST * 4] =
{
   /*  min, max, default, midi_chan  */
      0.0,   1.0,   0.5,  16,  /* HELIX_ARMLEN_ABS,  */
      0.0,   0.3,   0.0,   0,  /* HELIX_ARMLEN_REL,  */
     -0.5,   0.5,   0.0,  17,  /* HELIX_HEIGHT_ABS,  */
      0.0,  0.05,   0.02, 18,  /* HELIX_HEIGHT_REL,  */
      0.0,   0.3,   0.08, 19,  /* HELIX_SIZE_ABS,    */
     -0.1,   0.1,   0.0,   0,  /* HELIX_SIZE_REL,    */
      0.0, 180.0,   0.0,   0,  /* HELIX_ANGLE_ABS,   */
      0.0, 180.0,  20.0,  20,  /* HELIX_ANGLE_REL,   */
    -90.0,  90.0,   0.0,  21,  /* HELIX_ROLL_ABS,    */
    -90.0,  90.0,   0.0,   0,  /* HELIX_ROLL_REL,    */
    -90.0,  90.0,   0.0,  22,  /* HELIX_WAVE_ABS,    */
    -90.0,  90.0,   0.0,   0,  /* HELIX_WAVE_REL,    */
    -90.0,  90.0,   0.0,  23,  /* HELIX_WAG_ABS,     */
    -90.0,  90.0,   0.0,   0,  /* HELIX_WAG_REL,     */
      0.0, 100.0,  70.0,  27,  /* HELIX_L_ABS,       */
     -2.5,   2.5,  -1.5,  28,  /* HELIX_L_REL,       */
   -100.0, 100.0,   0.0,  29,  /* HELIX_A_ABS,       */
     -5.0,   5.0,   0.0,  30,  /* HELIX_A_REL,       */
   -100.0, 100.0, 100.0,  31,  /* HELIX_B_ABS,       */
     -5.0,   5.0,  -3.7,  33,  /* HELIX_B_REL,       */
      0.0,   1.0,   1.0,  24,  /* HELIX_OPACITY_ABS, */
      0.0,   1.0,   0.0,   0,  /* HELIX_OPACITY_REL, */
    -90.0,  90.0, -50.0,   1,  /* HELIX_VIEW_ROT_X,  */
    -90.0,  90.0,   0.0,   2,  /* HELIX_VIEW_ROT_Y,  */
    -90.0,  90.0,   0.0,   6,  /* HELIX_VIEW_ROT_Z,  */
     -1.0,   1.0,   0.0,  25,  /* HELIX_ROTATION_ABS */
     3.0,   16.0,   4.0,   0,  /* HELIX_N_VERTICES   */
};

static void
helix_elements_configure (HelixWindow *hwin)
{
  gint i;

  hwin->master_element->absolute = attr_zero;
  hwin->master_element->relative = attr_zero;
  hwin->master_element->absolute.vertices = 4;

  for (i = 0; i < hwin->nelements; i++)
    {
      HelixElement *elem = &hwin->elements[i];

      elem->absolute = hwin->master_element->absolute;
      elem->relative = hwin->master_element->relative;
    }
}


static void
helix_elements_propagate (HelixWindow *hwin)
{
  HelixElement *elem, *prev;
  gint i;

  for (i = hwin->nelements - 1; i >= 0; i--)
    {
      elem = &hwin->elements[i];

      if (i > 0)
        prev = &hwin->elements[i-1];
      else
        prev = hwin->master_element;

      /* copy behaviour from previous element */
      elem->absolute = prev->absolute;
      elem->relative = prev->relative;
    }

#if 0
  {
    static gint count = 0;
    gdouble time = hwin->ticks + hwin->tick_fraction;
    /* setup master element */

    count = (count + 1) % 250;

    elem = hwin->master_element;

    if (count < 200)
    {
      if (elem->relative.angle > -20)
        elem->relative.angle -= 1;
    }
    else
    {
      if (elem->relative.angle < 0)
        elem->relative.angle += 1;
    }

    elem->absolute.size = 0.06 + sin (time/G_PI * 2) * 0.01;
    elem->absolute.armlen = 0.4;
    elem->absolute.roll = 90;
    elem->absolute.wave = 90 + sin (time/G_PI*3) * 20;

    // elem->relative.wave = 90;

    elem->relative.armlen = 0.0075; // + sin (time/G_PI) * 0.05;
    // elem->relative.wag = 10 * sin (time/G_PI * 1.05);
    elem->relative.height = 1.5 / hwin->nelements;
  }
#endif
}


static void
helix_attribs_mix (gdouble        prev_factor,
                   HelixAttribs  *prev,
                   gdouble        this_factor,
                   HelixAttribs  *this,
                   HelixAttribs  *accu)
{
  gdouble d;

  accu->armlen  = prev_factor * prev->armlen  + this_factor * this->armlen;
  accu->wag     = prev_factor * prev->wag     + this_factor * this->wag;
  accu->height  = prev_factor * prev->height  + this_factor * this->height;
  accu->wave    = prev_factor * prev->wave    + this_factor * this->wave;
  accu->size    = prev_factor * prev->size    + this_factor * this->size;
  accu->roll    = prev_factor * prev->roll    + this_factor * this->roll;
  accu->opacity = prev_factor * prev->opacity + this_factor * this->opacity;

  accu->Lab[0]  = prev_factor * prev->Lab[0]  + this_factor * this->Lab[0];
  accu->Lab[1]  = prev_factor * prev->Lab[1]  + this_factor * this->Lab[1];
  accu->Lab[2]  = prev_factor * prev->Lab[2]  + this_factor * this->Lab[2];

  d = this->angle - prev->angle;
  accu->angle   = (prev_factor * (d > 180  ? prev->angle + 360 : prev->angle) +
                   this_factor * (d < -180 ? this->angle + 360 : this->angle));

  accu->vertices = (prev_factor * prev->vertices +
                    this_factor * this->vertices);
}


static int
helix_z_order_compare (const void *a,
                       const void *b)
{
  HelixIndexZ *a1 = (HelixIndexZ *) a;
  HelixIndexZ *b1 = (HelixIndexZ *) b;

  if (a1->z_depth < b1->z_depth)
    return -1;
  else if (a1->z_depth > b1->z_depth)
    return 1;
  else
    return 0;
}


/* draw a frame */
static gboolean
helix_frame_draw (GtkWidget      *widget,
                  GdkEventExpose *event,
                  gpointer        data)
{
  HelixWindow  *hwin = (HelixWindow *) data;
  HelixElement *prev;
  GLfloat color [4]     = { 0.0, 0.0, 0.0, 0.5 };
  GLfloat specular [4]  = { 0.5, 0.5, 0.5, 1.0 };
  GLfloat shininess [1] = { 70.0 };
  gint i, j;
  static gdouble total_height = 0;
  HelixIndexZ  *ordering_z = g_new0 (HelixIndexZ, hwin->nelements);
  HelixAttribs accu_rel = attr_zero;
  gdouble  rot_x, rot_y, rot_z;

  hwin->rotation_z += gtk_adjustment_get_value (hwin->adjustments[HELIX_ROTATION_REL]) * 360 / hwin->fps;

  rot_x = gtk_adjustment_get_value (hwin->adjustments[HELIX_VIEW_ROT_X]);
  rot_y = gtk_adjustment_get_value (hwin->adjustments[HELIX_VIEW_ROT_Y]);
  rot_z = gtk_adjustment_get_value (hwin->adjustments[HELIX_VIEW_ROT_Z]);

  glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, specular);
  glMaterialfv (GL_FRONT_AND_BACK, GL_SHININESS, shininess);

  glPushMatrix ();
  glRotatef (rot_x, 1.0, 0.0, 0.0);
  glRotatef (rot_y, 0.0, 1.0, 0.0);
  glRotatef (hwin->rotation_z + rot_z, 0.0, 0.0, 1.0);
  glTranslatef (0.0, 0.0, - total_height / 2.0);

  hwin->master_element->absolute.armlen  = gtk_adjustment_get_value (hwin->adjustments[HELIX_ARMLEN_ABS]);
  hwin->master_element->relative.armlen  = gtk_adjustment_get_value (hwin->adjustments[HELIX_ARMLEN_REL]);
  hwin->master_element->absolute.height  = gtk_adjustment_get_value (hwin->adjustments[HELIX_HEIGHT_ABS]);
  hwin->master_element->relative.height  = gtk_adjustment_get_value (hwin->adjustments[HELIX_HEIGHT_REL]);
  hwin->master_element->absolute.size    = gtk_adjustment_get_value (hwin->adjustments[HELIX_SIZE_ABS]);
  hwin->master_element->relative.size    = gtk_adjustment_get_value (hwin->adjustments[HELIX_SIZE_REL]);
  hwin->master_element->absolute.angle   = gtk_adjustment_get_value (hwin->adjustments[HELIX_ANGLE_ABS]);
  hwin->master_element->relative.angle   = gtk_adjustment_get_value (hwin->adjustments[HELIX_ANGLE_REL]);
  hwin->master_element->absolute.roll    = gtk_adjustment_get_value (hwin->adjustments[HELIX_ROLL_ABS]);
  hwin->master_element->relative.roll    = gtk_adjustment_get_value (hwin->adjustments[HELIX_ROLL_REL]);
  hwin->master_element->absolute.wave    = gtk_adjustment_get_value (hwin->adjustments[HELIX_WAVE_ABS]);
  hwin->master_element->relative.wave    = gtk_adjustment_get_value (hwin->adjustments[HELIX_WAVE_REL]);
  hwin->master_element->absolute.wag     = gtk_adjustment_get_value (hwin->adjustments[HELIX_WAG_ABS]);
  hwin->master_element->relative.wag     = gtk_adjustment_get_value (hwin->adjustments[HELIX_WAG_REL]);
  hwin->master_element->absolute.opacity = gtk_adjustment_get_value (hwin->adjustments[HELIX_OPACITY_ABS]);
  hwin->master_element->relative.opacity = gtk_adjustment_get_value (hwin->adjustments[HELIX_OPACITY_REL]);
  hwin->master_element->absolute.Lab[0]  = gtk_adjustment_get_value (hwin->adjustments[HELIX_L_ABS]);
  hwin->master_element->relative.Lab[0]  = gtk_adjustment_get_value (hwin->adjustments[HELIX_L_REL]);
  hwin->master_element->absolute.Lab[1]  = gtk_adjustment_get_value (hwin->adjustments[HELIX_A_ABS]);
  hwin->master_element->relative.Lab[1]  = gtk_adjustment_get_value (hwin->adjustments[HELIX_A_REL]);
  hwin->master_element->absolute.Lab[2]  = gtk_adjustment_get_value (hwin->adjustments[HELIX_B_ABS]);
  hwin->master_element->relative.Lab[2]  = gtk_adjustment_get_value (hwin->adjustments[HELIX_B_REL]);
  hwin->master_element->absolute.vertices = gtk_adjustment_get_value (hwin->adjustments[HELIX_N_VERTICES]);

  if (hwin->texture && hwin->textured)
    {
      glEnable (GL_TEXTURE_2D);
      glDisable (GL_DEPTH_TEST);
    }
  else
    {
      glDisable (GL_TEXTURE_2D);
      glEnable (GL_DEPTH_TEST);
    }

  /*
   * calculate the current state of the elements
   * and store their calculated transformation matrix
   */
  prev = hwin->master_element;

  for (j = 0; j < hwin->nelements; j++)
    {
      HelixElement *elem = &hwin->elements[j];
      HelixAttribs  curr_rel;

      gdouble dt = hwin->tick_fraction;

      /* interpolate the relative and absolute attributes */
      helix_attribs_mix (1.0 - dt, &elem->relative,
                         dt,       &prev->relative,
                         &curr_rel);

      helix_attribs_mix (1.0 - dt, &elem->absolute,
                         dt,       &prev->absolute,
                         &elem->current);

      /* accumulate the current relative attributes */
      helix_attribs_mix (1.0, &accu_rel,
                         1.0, &curr_rel,
                         &accu_rel);

      /*
       * the used values are the accumulated relative plus
       * the current absolute values.
       */
      helix_attribs_mix (1.0, &elem->current,
                         1.0, &accu_rel,
                         &elem->current);

      glPushMatrix ();

      glRotated (elem->current.angle, 0.0, 0.0, 1.0);
      glTranslated (0.0, elem->current.armlen, elem->current.height);
      glRotated (elem->current.roll, 0.0, 1.0, 0.0);
      glRotated (elem->current.wave, 1.0, 0.0, 0.0);
      /* wag = 0: flat side towards the center */
      glRotated (180 + elem->current.wag - 180.0 / elem->current.vertices,
                 0.0, 0.0, 1.0);
      glScaled (fabs (elem->current.size),
                fabs (elem->current.size),
                fabs (elem->current.size));

      /*
       * calculate the depth of the center of a element in world coords.
       * the center is (0,0,0,1)  (homogenous), thus
       * M * (0,0,0,1) =  (m[12], m[13], m[14], m[15])
       * and unhomogenizing gives z = m[14]/m[15].
       */
      glGetDoublev (GL_MODELVIEW_MATRIX, elem->transform);
      elem->z_depth = elem->transform[14] / elem->transform[15];

      ordering_z[j].index = j;
      ordering_z[j].z_depth = elem->z_depth;

      glPopMatrix ();

      prev = elem;
    }

  /*
   * sort the indicies of the elements according to their distance
   * to the viewer
   */

  qsort (ordering_z, hwin->nelements,
         sizeof (HelixIndexZ), helix_z_order_compare);

  /*
   * finally draw them to the screen in the correct order.
   */


  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  for (j = 0; j < hwin->nelements; j++)
    {
      HelixElement *elem = &hwin->elements[ordering_z[j].index];
      gdouble depth, opacity;
      gdouble X, Y, Z, R, G, B;

      glLoadMatrixd (elem->transform);

      depth = CLAMP ((elem->z_depth + 3.5), 0.0, 1.0);
      opacity = CLAMP (elem->current.opacity, 0.0, 1.0);

      if (hwin->texture && hwin->textured)
        {
          color[0] = color[1] = color[2] = 1.0;
          color[3] = opacity;
        }
      else
        {
          /*  as seen on http://www.easyrgb.com/math.php  */
          /*  L*a*b*  -->  XYZ  */

          Y = (elem->current.Lab[0] + 16) / 116;
          X = elem->current.Lab[1] / 500 + Y;
          Z = Y - elem->current.Lab[2] / 200;

          if (pow (Y, 3) > 0.008856)
            Y = pow (Y, 3);
          else
            Y = (Y - 16. / 116) / 7.787;

          if (pow (X, 3) > 0.008856)
            X = pow (X, 3);
          else
            X = (X - 16. / 116) / 7.787;

          if (pow (Z, 3) > 0.008856)
            Z = pow (Z, 3);
          else
            Z = (Z - 16. / 116) / 7.787;

          /* Observer= 2, Illuminant= D65  */
          X *=  95.047 / 100;
          Y *= 100.000 / 100;
          Z *= 108.883 / 100;

          /*  XYZ --> RGB  */

          R = X *  3.2406 + Y * -1.5372 + Z * -0.4986;
          G = X * -0.9689 + Y *  1.8758 + Z *  0.0415;
          B = X *  0.0557 + Y * -0.2040 + Z *  1.0570;

          if (R > 0.0031308)
            R = 1.055 * pow (R, 1.0 / 2.4) - 0.055;
          else
            R = 12.92 * R;

          if (G > 0.0031308)
            G = 1.055 * pow (G, 1.0 / 2.4) - 0.055;
          else
            G = 12.92 * G;

          if (B > 0.0031308)
            B = 1.055 * pow (B, 1.0 / 2.4) - 0.055;
          else
            B = 12.92 * B;

          color[0] = CLAMP (R, 0.0, 1.0);
          color[1] = CLAMP (G, 0.0, 1.0);
          color[2] = CLAMP (B, 0.0, 1.0);
          color[3] = opacity;
        }

      glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);

      glBegin(GL_POLYGON);
      /* clockwise to make the normal point into the correct direction */
      for (i = ceil (elem->current.vertices) - 1; i >= 0; i--)
        {
          gdouble x, y;

          x = sin (2.0 * G_PI * i / elem->current.vertices);
          y = cos (2.0 * G_PI * i / elem->current.vertices);

          if (hwin->texture && hwin->textured)
            glTexCoord2f (0.5 + x / 2, 0.5 + y / 2);

          glVertex3f (x, y, 0.0);
        }
      glEnd ();

#if 1
      color[0] = color[1] = color[2] = 1.0;
      color[3] = 1.0;

      glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
      glLineWidth (1.0);

      glBegin(GL_LINE_LOOP);
      /* clockwise to make the normal point into the correct direction */
      for (i = ceil (elem->current.vertices) - 1; i >= 0; i--)
        {
          glVertex3f (sin (2.0 * G_PI * i / elem->current.vertices),
                      cos (2.0 * G_PI * i / elem->current.vertices),
                      0.0);
        }
      glEnd ();
#endif
    }

  glPopMatrix ();

  total_height = accu_rel.height;

  g_free (ordering_z);

  return TRUE;
}


/* draw a stereo frame */
static gboolean
helix_frame_draw_stereo (GtkWidget      *widget,
                         GdkEventExpose *event,
                         gpointer        data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);
  gint width = widget->allocation.width;
  gint height = widget->allocation.height;

  GLfloat h = (GLfloat) (height) / (GLfloat) (width);
  GLfloat delta = 0.0 * 0.2;

  static gboolean cleared = FALSE;

  if (delta > 0.001)
    {
      width = width / 2;
      h = (GLfloat) (height) / (GLfloat) (width);
    }
  else
    {
      delta = 0.0;
    }

  /*** OpenGL BEGIN ***/
  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return FALSE;

  if (!cleared && (cleared = TRUE))
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

#if 0
  /* fade old picture */
  glMaterialfv (GL_FRONT, GL_AMBIENT_AND_DIFFUSE, color);
  glRectf (-5, -5, 5, 5);
  glClear (GL_DEPTH_BUFFER_BIT);
#else
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
#endif

  glViewport (0, 0, width, height);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  glFrustum (MIN (-.5, -.5 / h) - delta, MAX (.5, .5 / h) - delta,
             MIN (-.5, -.5 * h), MAX (.5, .5 * h), 2.0, 10.0);
  glTranslatef (-delta, 0, 0);

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  glTranslatef (0.0, 0.0, -3);

  helix_frame_draw (widget, event, data);

  if (delta > 0.001)
    {
      glViewport (width, 0, width, height);
      glMatrixMode (GL_PROJECTION);
      glLoadIdentity ();
      glFrustum (MIN (-.5, -.5 / h) + delta, MAX (.5, .5 / h) + delta,
                 MIN (-.5, -.5 * h), MAX (.5, .5 * h), 2.0, 10.0);
      glTranslatef (delta, 0, 0);

      glMatrixMode (GL_MODELVIEW);
      glLoadIdentity ();
      glTranslatef (0.0, 0.0, -3);

      helix_frame_draw (widget, event, data);
    }

  if (gdk_gl_drawable_is_double_buffered (gldrawable))
    gdk_gl_drawable_swap_buffers (gldrawable);
  else
    glFlush ();

  gdk_gl_drawable_gl_end (gldrawable);
  /*** OpenGL END ***/

  return TRUE;
}


static gboolean
helix_frame_next (HelixWindow *hwin)
{
  gdouble ticks;

  ticks = 7 * g_timer_elapsed (hwin->timer, NULL);

  while (ticks - hwin->ticks > 1.0)
    {
      helix_elements_propagate (hwin);
      hwin->ticks += 1;
    }

  hwin->tick_fraction = ticks - hwin->ticks;

  if (hwin->visibility)
    gtk_widget_queue_draw (hwin->drawing_area);

  return TRUE;
}


/* new window size or exposure */
static gboolean
helix_configure_event (GtkWidget         *widget,
                       GdkEventConfigure *event,
                       gpointer           data)
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  GLfloat h = (GLfloat) (widget->allocation.height) / (GLfloat) (widget->allocation.width);

  /*** OpenGL BEGIN ***/
  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return FALSE;

  glViewport (0, 0, widget->allocation.width, widget->allocation.height);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  glFrustum (MIN (-.5, -.5 / h), MAX (.5, .5 / h),
             MIN (-.5, -.5 * h), MAX (.5, .5 * h), 2.0, 10.0);

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();
  glTranslatef (0.0, 0.0, -3);

  gdk_gl_drawable_gl_end (gldrawable);
  /*** OpenGL END ***/

  return TRUE;
}


static gboolean
helix_visibility_update (GtkWidget   *widget,
                         GdkEventAny *event,
                         gpointer     data)
{
  HelixWindow *hwin = (HelixWindow *) data;

  switch (event->type)
    {
      case GDK_MAP:
          hwin->visibility = TRUE;
          break;
      case GDK_UNMAP:
          hwin->visibility = FALSE;
          break;
      case GDK_VISIBILITY_NOTIFY:
          hwin->visibility = ((GdkEventVisibility *) event)->state != GDK_VISIBILITY_FULLY_OBSCURED;
          break;
      default:
          g_printerr ("Unknown event in helix_visibility_update!\n");
          hwin->visibility = TRUE;
          break;
    }

  return TRUE;
}


/* change view angle, exit upon ESC */
static gboolean
helix_key_press_event (GtkWidget   *widget,
                       GdkEventKey *event,
                       gpointer     data)
{
  static gboolean fullscreen = FALSE;
  gint i;
  gdouble angle;
  HelixWindow *hwin = (HelixWindow *) data;

  switch (event->keyval)
    {
    case GDK_Up:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_X]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_X]),
                    "target-value", angle + 5.0, NULL);
      break;
    case GDK_Down:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_X]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_X]),
                    "target-value", angle - 5.0, NULL);
      break;
    case GDK_Left:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_Y]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_Y]),
                    "target-value", angle + 5.0, NULL);
      break;
    case GDK_Right:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_Y]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_Y]),
                    "target-value", angle - 5.0, NULL);
      break;
    case GDK_z:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_Z]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_Z]),
                    "target-value", angle + 5.0, NULL);
      break;
    case GDK_Z:
      g_object_get (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_Z]),
                    "target-value", &angle, NULL);
      g_object_set (G_OBJECT (hwin->adjustments[HELIX_VIEW_ROT_Z]),
                    "target-value", angle - 5.0, NULL);
      break;
    case GDK_F11:
    case GDK_f:
    case GDK_F:
      if (fullscreen)
        gtk_window_unfullscreen (GTK_WINDOW (hwin->window));
      else
        gtk_window_fullscreen (GTK_WINDOW (hwin->window));
      fullscreen = !fullscreen;
      break;
    case GDK_Escape:
    case GDK_q:
    case GDK_Q:
      gtk_main_quit ();
      break;
    case GDK_T:
    case GDK_t:
      hwin->textured = ! hwin->textured;
      break;
    case GDK_R:
    case GDK_r:
      for (i = 0; i < HELIX_SLOT_LAST; i++)
        {
          g_object_set (G_OBJECT (hwin->adjustments[i]),
                        "target-value", helix_defaults[i*4 + 2], NULL);
        }
      break;
    default:
      return FALSE;
    }

  return TRUE;
}


static void
helix_init_gl (GtkWidget *widget,
               gpointer   data)
{
  HelixWindow *hwin = (HelixWindow *) data;
  GdkGLContext *glcontext = gtk_widget_get_gl_context (widget);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (widget);

  static GLfloat pos[4] = {5.0, 5.0, 15.0, 0.0};
  static GLfloat white[4] = {1.0, 1.0, 1.0, 1.0};

  /*** OpenGL BEGIN ***/
  if (!gdk_gl_drawable_gl_begin (gldrawable, glcontext))
    return;

  if (hwin->texture)
    {
      glPixelStorei (GL_UNPACK_ALIGNMENT, 1);
      glTexImage2D (GL_TEXTURE_2D, 0, 4,
                    gdk_pixbuf_get_width (hwin->texture),
                    gdk_pixbuf_get_height (hwin->texture),
                    0, GL_RGBA,
                    GL_UNSIGNED_BYTE,
                    gdk_pixbuf_get_pixels (hwin->texture));
      glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
      glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
      glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
      glTexParameterf (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
      glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
      glShadeModel (GL_FLAT);
    }

  glLightfv (GL_LIGHT0, GL_POSITION, pos);
  glLightfv (GL_LIGHT0, GL_DIFFUSE, white);
  glLightfv (GL_LIGHT0, GL_SPECULAR, white);
  glEnable (GL_LIGHTING);
  glEnable (GL_LIGHT0);

  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable (GL_BLEND);
  glEnable (GL_LINE_SMOOTH);

  glEnable (GL_NORMALIZE);
  glLightModeli (GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);

  glEnable (GL_POLYGON_OFFSET_FILL);
  glPolygonOffset(1.0, 1.0);

#if 0
  g_printerr ("\n");
  g_printerr ("GL_RENDERER   = %s\n", (char *) glGetString (GL_RENDERER));
  g_printerr ("GL_VERSION    = %s\n", (char *) glGetString (GL_VERSION));
  g_printerr ("GL_VENDOR     = %s\n", (char *) glGetString (GL_VENDOR));
  g_printerr ("GL_EXTENSIONS = %s\n", (char *) glGetString (GL_EXTENSIONS));
  g_printerr ("\n");
#endif

  gdk_gl_drawable_gl_end (gldrawable);
  /*** OpenGL END ***/
}


static void
helix_setup_adjustments (HelixWindow *hwin)
{
  gint i;

  hwin->midi_source = g_object_new (MIDI_TYPE_SOURCE,
                                    "device", "alsa",
                                    "channel", -1,
                                    NULL);

  for (i = 0; i < HELIX_SLOT_LAST; i++)
    {
      gdouble lower, upper, value, step, page;
      gint controller;

      lower = helix_defaults[i*4 + 0];
      upper = helix_defaults[i*4 + 1];
      value = helix_defaults[i*4 + 2];

      controller = helix_defaults[i*4 + 3];

      step = (upper - lower) / 99;
      page = (upper - lower) / 11;

      hwin->adjustments[i] = g_object_new (MIDI_TYPE_ADJUSTMENT,
                                           "lower", lower,
                                           "upper", upper,
                                           "value", value,
                                           "target-value", value,
                                           "step-increment", step,
                                           "page-increment", page,
                                           "midi-source", hwin->midi_source,
                                           "controller", controller,
                                           NULL);
    }
}

HelixWindow *
helix_setup_ui ()
{
  GdkGLConfig *glconfig;
  GtkWidget *vbox;
  HelixWindow *hwin;

  hwin = g_new0 (HelixWindow, 1);

  hwin->texture = gdk_pixbuf_new_from_file ("texture.png", NULL);

  if (hwin->texture)
    {
      gboolean success = TRUE;

      success = success && gdk_pixbuf_get_has_alpha (hwin->texture);
      success = success && (gdk_pixbuf_get_n_channels (hwin->texture) == 4);
      success = success && (gdk_pixbuf_get_bits_per_sample (hwin->texture) == 8);

      success = success && (gdk_pixbuf_get_width (hwin->texture) *
                               gdk_pixbuf_get_n_channels (hwin->texture)
                            == gdk_pixbuf_get_rowstride (hwin->texture));

      if (!success)
        {
          g_object_unref (G_OBJECT (hwin->texture));
          hwin->texture = NULL;
        }
    }

  /* create timer */
  if (hwin->timer == NULL)
    hwin->timer = g_timer_new ();

  g_timer_start (hwin->timer);

  hwin->fps = 50;
  hwin->nelements = 54;
  hwin->elements = g_new0 (HelixElement, hwin->nelements);
  hwin->master_element = g_new0 (HelixElement, 1);

  helix_elements_configure (hwin);

  /*
   * Configure OpenGL-capable visual.
   */

  /* Try double-buffered visual */
  glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGB    |
                                        GDK_GL_MODE_DEPTH  |
                                        GDK_GL_MODE_DOUBLE);
  if (glconfig == NULL)
    {
      g_printerr ("*** Cannot find the double-buffered visual.\n");
      g_printerr ("*** Trying single-buffered visual.\n");

      /* Try single-buffered visual */
      glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGB   |
                                            GDK_GL_MODE_DEPTH);
      if (glconfig == NULL)
        {
          g_printerr ("*** No appropriate OpenGL-capable visual found.\n");
          exit (1);
        }
    }

  /*
   * Top-level window.
   */

  hwin->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (hwin->window), "Helix");

  /* Get automatically redrawn if any of their children changed allocation. */
  gtk_container_set_reallocate_redraws (GTK_CONTAINER (hwin->window), TRUE);

  g_signal_connect (G_OBJECT (hwin->window), "delete_event",
                    G_CALLBACK (gtk_main_quit), NULL);

  /*
   * VBox.
   */

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (hwin->window), vbox);
  gtk_widget_show (vbox);

  /*
   * Drawing area for drawing OpenGL scene.
   */

  hwin->drawing_area = gtk_drawing_area_new ();
  gtk_widget_set_size_request (hwin->drawing_area, 600, 600);

  /* Set OpenGL-capability to the widget. */
  gtk_widget_set_gl_capability (hwin->drawing_area,
                                glconfig,
                                NULL,
                                TRUE,
                                GDK_GL_RGBA_TYPE);

  gtk_widget_add_events (hwin->drawing_area,
                         GDK_VISIBILITY_NOTIFY_MASK);

  g_signal_connect_after (G_OBJECT (hwin->drawing_area), "realize",
                          G_CALLBACK (helix_init_gl), hwin);
  g_signal_connect (G_OBJECT (hwin->drawing_area), "configure_event",
                    G_CALLBACK (helix_configure_event), NULL);
  g_signal_connect (G_OBJECT (hwin->drawing_area), "expose_event",
                    G_CALLBACK (helix_frame_draw_stereo), hwin);

  g_signal_connect (G_OBJECT (hwin->drawing_area), "map_event",
                    G_CALLBACK (helix_visibility_update), hwin);
  g_signal_connect (G_OBJECT (hwin->drawing_area), "unmap_event",
                    G_CALLBACK (helix_visibility_update), hwin);
  g_signal_connect (G_OBJECT (hwin->drawing_area), "visibility_notify_event",
                    G_CALLBACK (helix_visibility_update), hwin);

  g_signal_connect (G_OBJECT (hwin->window), "key_press_event",
                    G_CALLBACK (helix_key_press_event), hwin);

  gtk_box_pack_start (GTK_BOX (vbox), hwin->drawing_area, TRUE, TRUE, 0);

  gtk_widget_show (hwin->drawing_area);

  /*
   * Show window.
   */

  gtk_widget_show (hwin->window);

  helix_setup_adjustments (hwin);

  /*
   * Timeout function to maintain framerate
   */
  hwin->timeout_id = g_timeout_add_full (GDK_PRIORITY_REDRAW,
                                         1000 / hwin->fps,
                                         (GSourceFunc) helix_frame_next,
                                         hwin,
                                         NULL);
  return hwin;
}


int
main (int   argc,
      char *argv[])
{
  HelixWindow     *hwin;
  GtkWidget       *panel;

  gtk_init (&argc, &argv);
  gtk_gl_init (&argc, &argv);

  hwin = helix_setup_ui ();

  if (argc > 1)
    panel = helix_create_control_panel (hwin);

  gtk_main ();

  return 0;
}
